/*
 * glcricket.c
 * David Moore <dcm@csail.mit.edu>
 *
 * Visualization tool for Crickets running the robust quad localization
 * algorithm.  Displays their positions and debugging information about
 * the network.
 *
 * TODO: This code is a horrible mess.  A new architecture with
 * event-driven parsing is currently being worked on.
 *
 *
 * Copyright (C) 2004  Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#define __USE_ISOC99 1
 
#include <stdio.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <png.h>


#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIDTH   1280
#define HEIGHT  960

/* stuff about our window grouped together */
typedef struct {
    Display *dpy;
    int screen;
    Window win;
    GLXContext ctx;
    XSetWindowAttributes attr;
    Bool fs;
    Bool doubleBuffered;
    XF86VidModeModeInfo deskMode;
    int x, y;
    unsigned int width, height;
    unsigned int depth;    
} GLWindow;

#define STAT_EMPTY      0x0
#define STAT_VALID      0x1
#define STAT_VISIBLE    0x2
#define STAT_OPTIMIZED  0x4
#define STAT_MARKED     0x8
#define STAT_WEAK       0x10
#define STAT_MOVING     0x20
#define MAX_NODES 40
#define MAX_QUADS 500
struct hist_pos {
    float x, y;
    struct timeval time;
};
struct node_pos {
    int stat, newstat;
    float x, y;
    float last_x, last_y;
    float ox, oy;
    int id[4];
    char * name;
    struct hist_pos * hist;
    int hist_size;
    int hist_num;
    int vel_set;
    struct timeval vel_time;
    float vel_x, vel_y;
} pos[MAX_NODES+1];

struct quad_t {
    int v[3];
} quads[MAX_QUADS];
struct quad_t quads_new[MAX_QUADS];
int num_quads;
int num_quads_new;

struct timeval pings[MAX_NODES+1][MAX_NODES+1];
int pings_d[MAX_NODES+1][MAX_NODES+1];
int pings_summary[MAX_NODES+1][MAX_NODES+1];
int pings_summary_new[MAX_NODES+1][MAX_NODES+1];
int mob_dist[MAX_NODES+1][MAX_NODES+1];
int mob_dist_new[MAX_NODES+1];
int max_nodes = 40;

int num_truth = 0;
struct truth_pos {
    float x, y;
    char * name;
};
struct truth_pos * truth = NULL;

float fov = 100.0;
float target_fov = 100.0;
float origin_x = 0.0;
float target_ox = 0.0;
float origin_y = 0.0;
float target_oy = 0.0;
float ui_rot = 0.0;
float comp_rot = 0.0;
float trot = 0.0;
float aspect;
int ui_flip = 1;
int comp_flip = 1;
int show_quads = 0;
int show_pings = 1;
int show_all_history = 0;
int auto_zoom = 1;
int show_vertex = -1;
int summary_complete = 1;
int picking_edge = 0;
int edge_v[2];
int edge_picked = 0;
float opt_err;

int input_version = 2;
int do_cones = 0;

int do_rot = 0;
int do_flip = 0;
char rotname[16];
char flipname[16];
int rot_node = 0;
int flip_node = 0;
float last_rot = 0;
float last_cross = 0;

int paused = 0;
int pause_on_loc = 0;

int do_smoothing = 0;

void process_line_v1(char * str);
void process_line_v2(char * str);

void (*process_line)(char *) = process_line_v2;

static struct board_names {
    int id[4];
    int crc;
    char * name;
} names[] = {
    { {0x09, 0x9e, 0x84, 0xdb}, 0, "01" },
    { {0x09, 0x9e, 0x8e, 0xad}, 0, "02" },
    { {0x09, 0x9e, 0x88, 0x83}, 0, "06" },
    { {0x09, 0x9e, 0x9b, 0x96}, 0, "07" },
    { {0x09, 0x9e, 0x91, 0x26}, 0, "08" },
    { {0x09, 0x9e, 0xa5, 0x45}, 0, "10" },
    { {0x09, 0x9e, 0xb0, 0x6d}, 0x81, "11" },
    { {0x09, 0x9e, 0x9a, 0x3c}, 0xdd, "12" },
    { {0x09, 0x9e, 0x98, 0x80}, 0, "13" },
    { {0x09, 0x9e, 0xae, 0x73}, 0, "14" },
    { {0x09, 0x9e, 0x85, 0x2f}, 0, "15" },
    { {0x09, 0x9e, 0x9c, 0x63}, 0x4b, "16" },
    { {0x09, 0x9e, 0x94, 0xe1}, 0xf1, "17" },
    { {0x09, 0x9e, 0x9a, 0x01}, 0x7a, "18" },
    { {0x09, 0x9e, 0x88, 0xaf}, 0xca, "19" },
    { {0x09, 0x9e, 0x96, 0x6a}, 0x60, "20" },
    { {0x09, 0x9e, 0xb2, 0x3a}, 0, "21" },
    { {0x09, 0x9e, 0x8b, 0xd9}, 0, "22" },
    { {0x09, 0x9e, 0x86, 0xb8}, 0, "23" },
    { {0x09, 0x9e, 0x8f, 0x3b}, 0, "24" },
    { {0x09, 0x9e, 0x86, 0xc6}, 0, "25" },

    { {0x0a, 0x3c, 0x67, 0xfb}, 0, "030" },
    { {0x0a, 0x3c, 0x35, 0x30}, 0, "034" },


    { {0x09, 0x9e, 0x87, 0xee}, 0, "C" },
    { {0x09, 0x9e, 0x93, 0xdc}, 0, "E" },
    { {0x09, 0x9e, 0x9b, 0x74}, 0, "G" },
    { {0x09, 0x9e, 0xaa, 0x1e}, 0, "H" },
    { {0x09, 0x9e, 0x88, 0x2a}, 0xcb, "B03" },
    { {0x09, 0x9e, 0x90, 0x34}, 0, "9AL" },
    { {0x09, 0x9e, 0x8b, 0x99}, 0, "B08" },
    { {0x09, 0x9e, 0x8c, 0xfb}, 0, "B09" },
    { {0x09, 0x9e, 0xb5, 0x6f}, 0, "B07" },
    { {0x09, 0x9e, 0xa6, 0x23}, 0, "6A" },
    { {0x09, 0x9e, 0xac, 0x1f}, 0, "14P" },

    { {0x0a, 0x9e, 0x88, 0x83}, 0, "06b" },
    { {0x0a, 0x9e, 0x91, 0x26}, 0, "08b" },
    { {0x0a, 0x9e, 0x84, 0xdb}, 0, "01b" },
    { {0x0a, 0x9e, 0x8e, 0xad}, 0, "02b" },
    { {0x0a, 0x9e, 0x9b, 0x96}, 0, "07b" },
    { {0x0a, 0x9e, 0xa5, 0x45}, 0, "10b" },
    { {0x0a, 0x9e, 0xb0, 0x6d}, 0x81, "11b" },
    { {0x0a, 0x9e, 0x9a, 0x3c}, 0xdd, "12b" },
    { {0x0a, 0x9e, 0xae, 0x73}, 0, "14b" },
    { {0x0a, 0x9e, 0x85, 0x2f}, 0, "15b" },
    { {0x0a, 0x9e, 0x9c, 0x63}, 0x4b, "16b" },
    { {0x0a, 0x9e, 0x9a, 0x01}, 0x7a, "18b" },
    { {0x0a, 0x9e, 0x88, 0xaf}, 0xca, "19b" },
    { {0x0a, 0x9e, 0xb2, 0x3a}, 0, "21b" },
    { {0x0a, 0x9e, 0x8b, 0xd9}, 0, "22b" },
    { {0x0b, 0x9e, 0x84, 0xdb}, 0, "01c" },
    { {0x0b, 0x9e, 0x8e, 0xad}, 0, "02c" },
    { {0x0b, 0x9e, 0x88, 0x83}, 0, "06c" },
    { {0x0b, 0x9e, 0xae, 0x73}, 0, "14c" },
    { {0x0b, 0x9e, 0x9a, 0x01}, 0x7a, "18c" },
    { {0x0b, 0x9e, 0x88, 0xaf}, 0xca, "19c" },
    { {0x0c, 0x9e, 0x84, 0xdb}, 0, "01d" },
};
#define NUM_NAMES   (sizeof(names)/sizeof(struct board_names))

/* attributes for a single buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
//    GLX_DEPTH_SIZE, 16,
    None};

/* attributes for a double buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListDbl[] = { GLX_RGBA, GLX_DOUBLEBUFFER, 
    GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_ALPHA_SIZE, 8,
    GLX_DEPTH_SIZE, 16,
    None };

GLWindow GLWin;
GLuint base;
GLvoid glPrintf(int font, const char * fmt, ...);

int isrecorded = 0;
struct timeval crickettime;
struct timeval starttime;

void getcrickettime(struct timeval * tv)
{
    if (isrecorded) {
        tv->tv_sec = crickettime.tv_sec;
        tv->tv_usec = crickettime.tv_usec;
    }
    else
        gettimeofday(tv, NULL);
}

unsigned char outbuf[WIDTH*HEIGHT*3];

int write_png(char * filename, int w, int h, int s, unsigned char * buf)
{
  FILE *f=fopen(filename,"wb");
  int i;
  png_structp png_ptr;
  png_infop info_ptr;
  png_byte * row_pointers[h];

  if (!f)
    return -1;

  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
    return -1;

  png_set_compression_level(png_ptr, 2);
  info_ptr = png_create_info_struct(png_ptr);
  if (!png_ptr)
    {
      png_destroy_write_struct(&png_ptr, NULL);
      return -1;
    }
  
  png_init_io(png_ptr,f);
  png_set_IHDR(png_ptr, info_ptr, w, h, 8, PNG_COLOR_TYPE_RGB,
               PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
               PNG_FILTER_TYPE_DEFAULT);
  
  for (i=0; i<h; i++)
    {
      row_pointers[i] = buf + (h-i-1)*s;
    }
  png_set_rows(png_ptr, info_ptr, row_pointers);
  png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
  png_write_end(png_ptr, info_ptr);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(f);
  return 0;
}


void write_frame(char * filename)
{
    glReadPixels(0, 0, GLWin.width, GLWin.height,
            GL_RGB, GL_UNSIGNED_BYTE, outbuf);
    write_png(filename, GLWin.width, GLWin.height, GLWin.width*3, outbuf);
}

/* function called when our window is resized (should only happen in window mode) */
void resizeGLScene(unsigned int width, unsigned int height)
{
    unsigned int size;
    if (height == 0)    /* Prevent A Divide By Zero If The Window Is Too Small */
        height = 1;

    size = (width<height)?width:height;
   // glViewport((width-size)/2, (height-size)/2, size, size);    /* Reset The Current Viewport And Perspective Transformation */
    glViewport(0, 0, width, height);    /* Reset The Current Viewport And Perspective Transformation */
    aspect = width/(float)height;
    //glMatrixMode(GL_PROJECTION);
    //glLoadIdentity();
    //gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    //glMatrixMode(GL_MODELVIEW);
}

GLvoid buildFont(GLvoid)
{
    XFontStruct *font1, *font2;
    base = glGenLists(192);
    font1 = XLoadQueryFont(GLWin.dpy, "-*-helvetica-bold-r-normal--34-*-*-*-p-*-iso8859-1");
    font2 = XLoadQueryFont(GLWin.dpy, "-*-courier-bold-r-normal--34-*-*-*-m-*-iso8859-1");
    if (font1 == NULL) {
        font1 = XLoadQueryFont(GLWin.dpy, "fixed");
        if (font1 == NULL) {
            fprintf(stderr, "Could not load font\n");
            exit(1);
        }
    }

    glXUseXFont(font1->fid, 32, 96, base);
    glXUseXFont(font2->fid, 32, 96, base+96);
    XFreeFont(GLWin.dpy, font1);
    XFreeFont(GLWin.dpy, font2);
}

GLvoid deleteFont(GLvoid)
{
    glDeleteLists(base, 96);
}

/* general OpenGL initialization function */
int initGL(GLvoid)
{
    //glShadeModel(GL_SMOOTH);
    glClearColor(0.0, 0.0, 0.0, 0.0);
//    glClearDepth(1.0f);
//    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//    glEnable(GL_POLYGON_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POINT_SMOOTH);
    //glDepthFunc(GL_LEQUAL);
    //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    /* we use resizeGLScene once to set up our initial perspective */
    buildFont();
    resizeGLScene(GLWin.width, GLWin.height);
    glFlush();
    return True;
}

void init_pos()
{
    int i;
    for (i=0; i<max_nodes; i++) {
        pos[i].stat = STAT_EMPTY;
        pos[i].newstat = pos[i].stat;
    }
    pos[max_nodes].stat = STAT_VALID | STAT_VISIBLE;
    pos[max_nodes].newstat = pos[max_nodes].stat;
    pos[max_nodes].x = 0;
    pos[max_nodes].y = 0;
    pos[max_nodes].last_x = 0;
    pos[max_nodes].last_y = 0;
    pos[max_nodes].name = "Unknown";

    num_quads = 0;
}

float sec_diff(struct timeval t1, struct timeval t2)
{
    return (t1.tv_sec-t2.tv_sec) + ((float)t1.tv_usec-(float)t2.tv_usec)/1000000.0;
}

void draw_arc(float x, float y, float r, float th1, float th2)
{
    float th;
    for (th=th1; th < th2; th += 0.02) {
        glVertex3f(x + r*cos(th),
                y + r*sin(th), 0.0);
    }
    glVertex3f(x + r*cos(th2),
            y + r*sin(th2), 0.0);
}

void draw_node_ping(int i, int j, struct timeval * pt, int meas)
{
    struct timeval tv;
    float diff, dist, xdif, ydif, dist2;
    float as = 4.0;
    float spc = 6.5;

    getcrickettime(&tv);
    if (!(pos[j].stat & STAT_VISIBLE) ||
            !(pos[i].stat & STAT_VISIBLE) || meas == 0)
        return;
    if (pt) {
        if (pt->tv_sec == 0)
            return;
        diff = sec_diff(tv, *pt);
        if (diff >= 1.0) {
            pt->tv_sec = 0;
            return;
        }
        glColor4f(1.0, 1.0, 0.0, 1.0-diff);
    }
    else if (meas == 1) {
        glColor4f(1.0, 0.0, 0.0, 1.0);
    }
    else {
        glColor4f(0.0, 1.0, 0.0, 1.0);
    }
    glLineWidth(4.0);
    xdif = pos[j].x - pos[i].x;
    ydif = pos[j].y - pos[i].y;
    dist = sqrt(xdif*xdif+ydif*ydif);
    dist2 = meas/30.0;
    if (dist2-dist > -spc)
        dist2 = dist - spc;
//        glVertex3f(pos[j].x - xdif*4.0/dist - ydif*2.0/dist,
//                pos[j].y - ydif*4.0/dist + xdif*2.0/dist, 0.0);
    if (dist2 > spc) {
        glBegin(GL_LINE_STRIP);
        glVertex3f(pos[i].x + xdif*dist2/dist - ydif*as/dist,
                pos[i].y + ydif*dist2/dist + xdif*as/dist, 0.0);
        glVertex3f(pos[i].x + xdif*spc/dist - ydif*as/dist,
                pos[i].y + ydif*spc/dist + xdif*as/dist, 0.0);
        glEnd();
        glBegin(GL_TRIANGLES);
        glVertex3f(pos[i].x + xdif*spc/dist - ydif*as/dist,
                pos[i].y + ydif*spc/dist + xdif*as/dist, 0.0);
        glVertex3f(pos[i].x + xdif*(spc+2*as)/dist - ydif*2*as/dist,
                pos[i].y + ydif*(spc+2*as)/dist + xdif*2*as/dist, 0.0);
        glVertex3f(pos[i].x + xdif*(spc+2*as)/dist - ydif*as/dist,
                pos[i].y + ydif*(spc+2*as)/dist + xdif*as/dist, 0.0);
        glEnd();
    }

    dist2 = meas/30.0;
    if (fabs(dist2-dist) > spc) {
        float sig = 1;
        dist2 -= dist;
        if (dist2 < -spc) {
            glLineWidth(4.0);
            glLineStipple(1, 0xff00);
            glEnable(GL_LINE_STIPPLE);
            sig = -1;
        }
        glBegin(GL_LINE_STRIP);
        glVertex3f(pos[j].x + xdif*dist2/dist - ydif*as/dist,
                pos[j].y + ydif*dist2/dist + xdif*as/dist, 0.0);
        glVertex3f(pos[j].x + sig*xdif*spc/dist - ydif*as/dist,
                pos[j].y + sig*ydif*spc/dist + xdif*as/dist, 0.0);
        glEnd();
        if (dist2 < -spc)
            glDisable(GL_LINE_STIPPLE);
    }
}

void draw_node_vel(int i)
{
    if (!pos[i].vel_set)
        return;

    glLineWidth(4.0);
    glBegin(GL_LINES);
    glColor3f(1.0, 0.0, 0.0);
    glVertex3f(pos[i].x, pos[i].y, 0.0);
    glVertex3f(pos[i].x + pos[i].vel_x, pos[i].y + pos[i].vel_y, 0.0);
    glEnd();
    
}

void draw_mob_dist()
{
    int i, j;

    for (i=0; i<=max_nodes; i++) {
        if (!(pos[i].stat & STAT_MOVING) || !(pos[i].stat & STAT_VISIBLE))
            continue;
        for (j=0; j<=max_nodes; j++) {
            if (i != j && mob_dist[i][j] != 0)
                draw_node_ping(j, i, NULL, mob_dist[i][j]);
        }
    }
}

void draw_pings()
{
    int i, j;

    if (show_vertex != -1) {
        for (j=0; j<=max_nodes; j++) {
            if (j != show_vertex) {
                draw_node_ping(show_vertex, j, NULL,
                        pings_summary[show_vertex][j]);
                draw_node_ping(j, show_vertex, NULL,
                        pings_summary[j][show_vertex]);
            }
        }
        return;
    }

    for (i=0; i<=max_nodes; i++) {
        if (!(pos[i].stat & STAT_VISIBLE))
            continue;
        for (j=0; j<=max_nodes; j++) {
            if (j != i)
                draw_node_ping(i, j, &pings[i][j], pings_d[i][j]);
        }
    }
}

void draw_edge(int a, int b)
{
    glVertex3f(pos[a].x, pos[a].y, 0.0);
    glVertex3f(pos[b].x, pos[b].y, 0.0);
}

#define square(a)   ((a)*(a))

int get_normalized_dist(int i, int j)
{
    if (pos[i].stat & STAT_MOVING)
        return pings_summary[j][i];
    if (pos[j].stat & STAT_MOVING)
        return pings_summary[i][j];

    return (pings_summary[i][j] + pings_summary[j][i]) >> 1;
}

void trilaterate1(int i1, int i2, int i3, int i,
        float * x, float * y)
{
    float dsq, gam, xa, ya, xb, yb, xt1, xt2, yt1, yt2, d1, d2;
    float x1, y1, x2, y2, x3, y3;
    int r1, r2, r3;

    x1 = pos[i1].x*30;
    y1 = pos[i1].y*30;

    x2 = pos[i2].x*30;
    y2 = pos[i2].y*30;

    x3 = pos[i3].x*30;
    y3 = pos[i3].y*30;

    /*
    r1 = (pings_summary[i1][i] + pings_summary[i][i1]) >> 1;
    r2 = (pings_summary[i2][i] + pings_summary[i][i2]) >> 1;
    r3 = (pings_summary[i3][i] + pings_summary[i][i3]) >> 1;
    */
    r1 = get_normalized_dist(i1, i);
    r2 = get_normalized_dist(i2, i);
    r3 = get_normalized_dist(i3, i);

    dsq = square(x2-x1) + square(y2-y1);
    gam = sqrt((square((float)r2+r1) - dsq)*(dsq - square((float)r2-r1)));
    xa = -(square(r2)-square(r1))*(x2-x1)/(2.0*dsq) + (x1+x2)*0.5;
    ya = -(square(r2)-square(r1))*(y2-y1)/(2.0*dsq) + (y1+y2)*0.5;
    xb = (y2-y1)*gam/(2.0*dsq);
    yb = (x2-x1)*gam/(2.0*dsq);
    xt1 = xa - xb;
    xt2 = xa + xb;
    yt1 = ya + yb;
    yt2 = ya - yb;
    
    d1 = sqrt(square(xt1-x3) + square(yt1-y3));
    d2 = sqrt(square(xt2-x3) + square(yt2-y3));
    if (fabs(d1-r3) < fabs(d2-r3)) {
        *x = xt1/30;
        *y = yt1/30;
    }
    else {
        *x = xt2/30;
        *y = yt2/30;
    }
}

double rigidity(int a, int b, int c)
{
    double min, d, e;
    double costh;

    if (a <= b && a <= c)
        min = a, d = b, e = c;
    else if (b <= a && b <= c)
        min = b, d = a, e = c;
    else
        min = c, d = a, e = b;

    costh = (square(d)+square(e)-square(min))/(2.0*d*e);
    return min*(1-square(costh));
}

void print_rigidity(int vertex)
{
    int i, j, k, l;
    int dij, dik, dil, djk, dkl, dlj;

    for (i=0; i<num_quads; i++) {
        if (vertex == quads[i].v[0] || vertex == quads[i].v[1] ||
                vertex == quads[i].v[2])
            break;
    }
    if (i == num_quads)
        return;

    j = quads[i].v[0];
    k = quads[i].v[1];
    l = quads[i].v[2];
    i = max_nodes;

    dij = (pings_summary[j][i] + pings_summary[i][j]) >> 1;
    dik = (pings_summary[k][i] + pings_summary[i][k]) >> 1;
    dil = (pings_summary[l][i] + pings_summary[i][l]) >> 1;
    dkl = (pings_summary[l][k] + pings_summary[k][l]) >> 1;
    djk = (pings_summary[j][k] + pings_summary[k][j]) >> 1;
    dlj = (pings_summary[j][l] + pings_summary[l][j]) >> 1;

    printf("Rigidity of Quad %d %d %d %d: %.1f %.1f %.1f %.1f\n",
            i, j, k, l, rigidity(dij,dik,djk), rigidity(dij,dil,dlj),
            rigidity(dik,dil,dkl), rigidity(djk,dkl,dlj));
    printf("(%d %d %d) (%d %d %d) (%d %d %d) (%d %d %d)\n",
            dij, dik, djk, dij, dil, dlj, dik, dil, dkl, djk, dkl, dlj);
}

void draw_quads()
{
    int i;

    glColor4f(0.0, 0.0, 1.0, 1.0);
    glLineWidth(4.0);
    for (i=0; i<num_quads; i++) {
        if (show_vertex != -1 && show_vertex != quads[i].v[0] &&
                    show_vertex != quads[i].v[1] && show_vertex != quads[i].v[2])
            continue;
        glBegin(GL_LINES);
        draw_edge(quads[i].v[1], quads[i].v[2]);
        draw_edge(quads[i].v[0], quads[i].v[2]);
        draw_edge(max_nodes, quads[i].v[2]);
        glEnd();
        if (show_vertex != -1)
            glLineWidth(8.0);
        glBegin(GL_LINES);
        draw_edge(quads[i].v[0], quads[i].v[1]);
        draw_edge(max_nodes, quads[i].v[0]);
        draw_edge(max_nodes, quads[i].v[1]);
        glEnd();
        if (show_vertex != -1)
            break;
    }

    /* Draw the three possible trilaterated points */
    if (i != num_quads && show_vertex == quads[i].v[2]) {
        float x, y;
        glLineWidth(4.0);
        glColor4f(1.0, 1.0, 0.0, 1.0);
        glBegin(GL_LINES);
        trilaterate1(max_nodes, quads[i].v[0], quads[i].v[1], show_vertex,
                &x, &y);
        glVertex3f(x-4.0, y, 0.0);
        glVertex3f(x+4.0, y, 0.0);
        glVertex3f(x, y-4.0, 0.0);
        glVertex3f(x, y+4.0, 0.0);
        trilaterate1(max_nodes, quads[i].v[1], quads[i].v[0], show_vertex,
                &x, &y);
        glVertex3f(x-4.0, y, 0.0);
        glVertex3f(x+4.0, y, 0.0);
        glVertex3f(x, y-4.0, 0.0);
        glVertex3f(x, y+4.0, 0.0);
        trilaterate1(quads[i].v[0], quads[i].v[1], max_nodes, show_vertex,
                &x, &y);
        glVertex3f(x-4.0, y, 0.0);
        glVertex3f(x+4.0, y, 0.0);
        glVertex3f(x, y-4.0, 0.0);
        glVertex3f(x, y+4.0, 0.0);
        glEnd();
    }
}

void draw_circle(float x, float y, float z, float r)
{
    float th;
    for (th=0; th < 2*M_PI; th += 0.4) {
        glVertex3f(x + r*cos(th), y + r*sin(th), z);
    }
}

void draw_truth()
{
    int i, match=-1;
    float ox = 0;
    float oy = 0;

    for (i=0; i<num_truth; i++) {
        if (!strcmp(truth[i].name, pos[max_nodes].name)) {
            ox = -truth[i].x;
            oy = -truth[i].y;
            match = i;
            break;
        }
    }
    for (i=0; i<num_truth; i++) {
        if (i == match)
            continue;
        glBegin(GL_POLYGON);
        glColor4f(0.0, 1.0, 0.0, 0.5);
        draw_circle(truth[i].x+ox, truth[i].y+oy, 0.0, 3.0);
        glEnd();
        glRasterPos2f(truth[i].x+ox+3.5*cos(trot*M_PI/180),
                truth[i].y+oy-3.5*sin(trot*M_PI/180));
        glPrintf(0, "%s", truth[i].name);
    }
}

void draw_history()
{
    int i, j;
    float th = 1.0;
    struct timeval tv;
    float exp = 60.0;
    getcrickettime(&tv);
    glClearDepth(0.0);
    glDepthFunc(GL_NOTEQUAL);
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    for (i=0; i<=max_nodes; i++) {
        float x1, y1, x2, y2, xdif, ydif, dist, t1, t2;
        if (pos[i].hist_num == 0)
            continue;

        glClear(GL_DEPTH_BUFFER_BIT);
        glColor3f(0.5, 0.5, 0.92);
        x2 = pos[i].x;
        y2 = pos[i].y;
        t2 = 0;
        for (j=pos[i].hist_num-1; j>=0; j--) {
            x1 = pos[i].hist[j].x;
            y1 = pos[i].hist[j].y;
            if (show_all_history)
                t1 = 0;
            else
                t1 = sec_diff(tv, pos[i].hist[j].time);
            ydif = y2 - y1;
            xdif = x2 - x1;
            dist = hypot(ydif, xdif);
            if (t2 > exp) {
                break;
            }
            else if (t1 > exp) {
                x1 = x2 - xdif * (exp-t2)/(t1-t2);
                y1 = y2 - ydif * (exp-t2)/(t1-t2);
            }
            glBegin(GL_QUADS);
            glColor4f(0.5, 0.5, 0.92, (exp-t2)/exp);
            glVertex3f(x2 - ydif*th/dist, y2 + xdif*th/dist, 1.0);
            glVertex3f(x2 + ydif*th/dist, y2 - xdif*th/dist, 1.0);
            glColor4f(0.5, 0.5, 0.92, (exp-t1)/exp);
            glVertex3f(x1 + ydif*th/dist, y1 - xdif*th/dist, 1.0);
            glVertex3f(x1 - ydif*th/dist, y1 + xdif*th/dist, 1.0);
            glEnd();
            glBegin(GL_POLYGON);
            glColor4f(0.5, 0.5, 0.92, (exp-t1)/exp);
            draw_circle(x1, y1, 1.0, th);
            glEnd();
            x2 = x1;
            y2 = y1;
            t2 = t1;
        }
    }
    glDisable(GL_DEPTH_TEST);
}

void draw_nodes()
{
    int i;

    if (show_quads)
        draw_quads();
    for (i=0; i<=max_nodes; i++) {
        float x, y;
        glBegin(GL_POLYGON);
        if (i == max_nodes) {
            glColor3f(0.9, 0.1, 0.1);
        }
        else if (pos[i].stat & STAT_VISIBLE) {
            if (pos[i].stat & STAT_WEAK)
                glColor3f(0.92, 0.5, 0.5);
            else
                glColor3f(0.92, 0.92, 0.92);
        }
        else {
            continue;
        }
        x = pos[i].last_x;
        y = pos[i].last_y;
        //printf("x = %f y = %f\n", x, y);
        if ((pos[i].stat & STAT_MOVING) && pos[i].vel_set && do_smoothing) {
            float t;
            struct timeval tv;

            getcrickettime(&tv);
            t = sec_diff(tv, pos[i].vel_time);
            x += pos[i].vel_x * t;
            y += pos[i].vel_y * t;
            pos[i].x = x;
            pos[i].y = y;
            //printf("draw x = %f y = %f t = %f\n", x, y, t);
        }
        draw_circle(x, y, 0.0, 5.0);
        glEnd();
        if (pos[i].stat & STAT_MOVING) {
            glBegin(GL_POLYGON);
            glColor3f(0.0, 0.0, 0.0);
            draw_circle(x, y, 0.0, 1.6);
            glEnd();
        }
        if (pos[i].stat & STAT_OPTIMIZED) {
            glBegin(GL_POLYGON);
            glColor4f(0.0, 0.0, 1.0, 0.5);
            draw_circle(pos[i].ox, pos[i].oy, 0.0, 5.0);
            glEnd();
        }
        glColor3f(0.92, 0.92, 0.92);
        glRasterPos2f(x+ui_flip*comp_flip*6*sin((ui_rot+comp_rot)*M_PI/180),
                y+6*cos((ui_rot+comp_rot)*M_PI/180));
        glPrintf(0, "%s", pos[i].name);
      //  if (pos[i].stat & STAT_MOVING)
      //      draw_node_vel(i);
    }
    if (show_pings == 1)
        draw_pings();
    else if (show_pings == 2)
        draw_mob_dist();
}

#define MAX(a,b) ((a)>(b)?(a):(b))

void recompute_bounding(float * ox, float * oy, float * fov,
        float cox, float coy, float cfov)
{
    float minx = HUGE_VAL;
    float maxx = -HUGE_VAL;
    float miny = HUGE_VAL;
    float maxy = -HUGE_VAL;
    float x, y;
    float rot;
    float new_fov;
    int i, refresh = 0;

    for (i=0; i<=max_nodes; i++) {
        if (i != max_nodes && !(pos[i].stat & STAT_VISIBLE))
            continue;

        x = ui_flip * comp_flip * pos[i].x;
        rot = ui_rot + comp_rot;
        y = x * sin(rot*M_PI/180) + pos[i].y * cos(rot*M_PI/180);
        x = x * cos(rot*M_PI/180) - pos[i].y * sin(rot*M_PI/180);
        if (y > maxy)
            maxy = y;
        if (y < miny)
            miny = y;
        if (x > maxx)
            maxx = x;
        if (x < minx)
            minx = x;
    }
    maxy += 20;
    miny -= 20;
    maxx += 20;
    minx -= 20;

    /*
    if (maxx + *ox > *fov || minx + *ox < -*fov)
        refresh = 1;
    if (maxy + *oy > *fov/aspect || miny + *oy < -*fov/aspect)
        refresh = 1;

    if (!refresh)
        return;
*/
    new_fov = MAX((maxx - minx) / 2, (maxy - miny) * aspect / 2);
    if (((new_fov/cfov) > 1.1 || (new_fov/cfov) < 0.9) && new_fov > 100)
        *fov = new_fov;

    maxx = -(minx + maxx) / 2;
    if (fabs(maxx - cox) > 10)
        *ox = maxx;
    maxy = -(miny + maxy) / 2;
    if (fabs(maxy - coy) > 10)
        *oy = maxy;
}

void draw_grid()
{
    float y;
    glLineWidth(4.0);
    glBegin(GL_LINES);
    glColor4f(0.2, 0.2, 0.2, 1.0);
    y = ceil((-fov/aspect-origin_y)/50.0)*50.0;
    for ( ; y < fov/aspect-origin_y; y += 50) {
        glVertex3f(-fov-origin_x, y, 0.0);
        glVertex3f(fov-origin_x, y, 0.0);
    }
    y = ceil((-fov-origin_x)/50.0)*50.0;
    for ( ; y < fov-origin_x; y += 50) {
        glVertex3f(y, -fov/aspect-origin_y, 0.0);
        glVertex3f(y, fov/aspect-origin_y, 0.0);
    }
    glEnd();
}

#define EDGE_HIST 200
struct edge_len {
    float len, loc_len;
} edge[EDGE_HIST];
int edge_head = 0;
int edge_tail = 0;
struct timeval edgeupdate;

int new_edge = 0;
float new_edge_len;

void draw_edge_over_time()
{
    int i, c=0;
    float max = 0.1;
    float last_loc = 0.0;
    struct timeval tv;

    getcrickettime(&tv);
    if (tv.tv_sec > edgeupdate.tv_sec ||
            (tv.tv_sec == edgeupdate.tv_sec &&
             tv.tv_usec > edgeupdate.tv_usec)) {
        edge_head++;
        if (edge_head == EDGE_HIST)
            edge_head = 0;
        edge[edge_head].len = 0;
        edge[edge_head].loc_len = 0;
        
        if (new_edge) {
            edge[edge_head].len = new_edge_len;
        }
        new_edge = 0;

        if ((pos[edge_v[0]].stat & STAT_VISIBLE) &&
                (pos[edge_v[1]].stat & STAT_VISIBLE)) {
            edge[edge_head].loc_len =
                sqrt(square(pos[edge_v[0]].x - pos[edge_v[1]].x) + 
                        square(pos[edge_v[0]].y - pos[edge_v[1]].y));
        }

        if (edge_head == edge_tail) {
            edge_tail++;
            if (edge_tail == EDGE_HIST)
                edge_tail = 0;
        }

        edgeupdate.tv_sec += (edgeupdate.tv_usec+250000) / 1000000;
        edgeupdate.tv_usec = (edgeupdate.tv_usec+250000) % 1000000;
    }
    
    i = edge_head;
    while (i != edge_tail) {
        if (edge[i].len > max)
            max = edge[i].len;
        if (edge[i].loc_len > max)
            max = edge[i].loc_len;

        if (i==0)
            i = EDGE_HIST;
        i--;
    }

    glLineWidth(1.0);
    glBegin(GL_LINES);
    i = edge_head;
    while (i != edge_tail) {
        float x = -c/(float)EDGE_HIST;
        float lx = -(c-1)/(float)EDGE_HIST;

        if (edge[i].loc_len > 0.0 && last_loc > 0.0) {
            glColor4f(1.0, 1.0, 1.0, 1.0);
            glVertex3f(x, -0.95 + 0.20*edge[i].loc_len/max, 0.0);
            glVertex3f(lx, -0.95 + 0.20*last_loc/max, 0.0);
        }
        last_loc = edge[i].loc_len;
        if (edge[i].len > 0.0) {
            if (edge_v[0] == max_nodes)
                glColor4f(1.0, 1.0, 0.0, 1.0);
            else
                glColor4f(0.0, 1.0, 0.0, 1.0);
            glVertex3f(x, -0.95, 0.0);
            glVertex3f(x, -0.95 + 0.20*edge[i].len/max, 0.0);
        }

        if (i==0)
            i = EDGE_HIST;
        i--;
        c++;
    }
    glEnd();
}

/* Here goes our drawing code */
int drawGLScene(GLvoid)
{
    float m[] = {ui_flip, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
    float m2[] = {comp_flip, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
    float t;
    struct timeval tv;
    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    if (edge_picked) {
        draw_edge_over_time();
    }
    glScalef(1.0/fov, 1.0*aspect/fov, 1.0);
    glTranslatef(origin_x, origin_y, 0);
    draw_grid();
    
    if (truth) {
        glPushMatrix();
        glRotatef(trot, 0.0, 0.0, 1.0);
        draw_truth();
        glPopMatrix();
    }

    glRotatef(ui_rot, 0.0, 0.0, 1.0);
    glPushMatrix();
    glMultMatrixf(m);
    draw_history();
    glPopMatrix();
    glRotatef(comp_rot, 0.0, 0.0, 1.0);
    glMultMatrixf(m);
    glMultMatrixf(m2);
    draw_nodes();

    glLoadIdentity();
    glTranslatef(-1.0, 1.0, 0.0);
    glScalef(2.0/GLWin.width, -2.0/GLWin.height, 1.0);
    glColor3f(0.92, 0.92, 0.92);
    glRasterPos2f(10, 30);
    getcrickettime(&tv);
    t = sec_diff(tv, starttime);
    glPrintf(1, "t = %7.2f", t);
    
    if (GLWin.doubleBuffered)
    {
        glXSwapBuffers(GLWin.dpy, GLWin.win);
    }
    return True;    
}

int find_node_by_xy(int x, int y)
{
    float nx, ny, ox;
    int i, closest;
    float dist, cdist = HUGE_VAL;

    nx = (x*2/(float)GLWin.width - 1.0) * fov - origin_x;
    ny = -(y*2/(float)GLWin.height - 1.0) * fov/aspect - origin_y;

    ox = nx;
    nx = ox*cos(-(ui_rot+comp_rot)*M_PI/180) - ny*sin(-(ui_rot+comp_rot)*M_PI/180);
    ny = ox*sin(-(ui_rot+comp_rot)*M_PI/180) + ny*cos(-(ui_rot+comp_rot)*M_PI/180);
    nx = ui_flip * comp_flip * nx;

    
    for (i=0; i<=max_nodes; i++) {
        if (i != max_nodes && !(pos[i].stat & STAT_VISIBLE))
            continue;

        dist = hypot(pos[i].x - nx, pos[i].y - ny);
        if (dist < cdist) {
            cdist = dist;
            closest = i;
        }
    }

    if (cdist < 10.0)
        return closest;

    return -1;
}

/* function to release/destroy our resources and restoring the old desktop */
GLvoid killGLWindow(GLvoid)
{
    if (GLWin.ctx)
    {
        if (!glXMakeCurrent(GLWin.dpy, None, NULL))
        {
            printf("Could not release drawing context.\n");
        }
        glXDestroyContext(GLWin.dpy, GLWin.ctx);
        GLWin.ctx = NULL;
    }
    /* switch back to original desktop resolution if we were in fs */
    if (GLWin.fs)
    {
        XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, &GLWin.deskMode);
        XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
    }
    deleteFont();
    XCloseDisplay(GLWin.dpy);
}

/* this function creates our window and sets it up properly */
/* FIXME: bits is currently unused */
Bool createGLWindow(char* title, int width, int height, int bits,
                    Bool fullscreenflag)
{
    XVisualInfo *vi;
    Colormap cmap;
    int dpyWidth, dpyHeight;
    int i;
    int glxMajorVersion, glxMinorVersion;
    int vidModeMajorVersion, vidModeMinorVersion;
    XF86VidModeModeInfo **modes;
    int modeNum;
    int bestMode;
    Atom wmDelete;
    Window winDummy;
    unsigned int borderDummy;
    
    GLWin.fs = fullscreenflag;
    /* set best mode to current */
    bestMode = 0;
    /* get a connection */
    GLWin.dpy = XOpenDisplay(0);
    GLWin.screen = DefaultScreen(GLWin.dpy);
    XF86VidModeQueryVersion(GLWin.dpy, &vidModeMajorVersion,
        &vidModeMinorVersion);
    printf("XF86VidModeExtension-Version %d.%d\n", vidModeMajorVersion,
        vidModeMinorVersion);
    XF86VidModeGetAllModeLines(GLWin.dpy, GLWin.screen, &modeNum, &modes);
    /* save desktop-resolution before switching modes */
    GLWin.deskMode = *modes[0];
    /* look for mode with requested resolution */
    for (i = 0; i < modeNum; i++)
    {
        if ((modes[i]->hdisplay == width) && (modes[i]->vdisplay == height))
        {
            bestMode = i;
        }
    }
    /* get an appropriate visual */
    vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListDbl);
    if (vi == NULL)
    {
        vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListSgl);
        GLWin.doubleBuffered = False;
        printf("Only Singlebuffered Visual!\n");
    }
    else
    {
        GLWin.doubleBuffered = True;
        printf("Got Doublebuffered Visual!\n");
    }
    glXQueryVersion(GLWin.dpy, &glxMajorVersion, &glxMinorVersion);
    printf("glX-Version %d.%d\n", glxMajorVersion, glxMinorVersion);
    /* create a GLX context */
    GLWin.ctx = glXCreateContext(GLWin.dpy, vi, 0, GL_TRUE);
    /* create a color map */
    cmap = XCreateColormap(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        vi->visual, AllocNone);
    GLWin.attr.colormap = cmap;
    GLWin.attr.border_pixel = 0;

    if (GLWin.fs)
    {
        XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, modes[bestMode]);
        XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
        dpyWidth = modes[bestMode]->hdisplay;
        dpyHeight = modes[bestMode]->vdisplay;
        printf("Resolution %dx%d\n", dpyWidth, dpyHeight);
        XFree(modes);
    
        /* create a fullscreen window */
        GLWin.attr.override_redirect = True;
        GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
            0, 0, dpyWidth, dpyHeight, 0, vi->depth, InputOutput, vi->visual,
            CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
            &GLWin.attr);
        XWarpPointer(GLWin.dpy, None, GLWin.win, 0, 0, 0, 0, 0, 0);
		XMapRaised(GLWin.dpy, GLWin.win);
        XGrabKeyboard(GLWin.dpy, GLWin.win, True, GrabModeAsync,
            GrabModeAsync, CurrentTime);
        XGrabPointer(GLWin.dpy, GLWin.win, True, ButtonPressMask,
            GrabModeAsync, GrabModeAsync, GLWin.win, None, CurrentTime);
    }
    else
    {
        /* create a window in window mode*/
        GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
            0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
            CWBorderPixel | CWColormap | CWEventMask, &GLWin.attr);
        /* only set window title and handle wm_delete_events if in windowed mode */
        wmDelete = XInternAtom(GLWin.dpy, "WM_DELETE_WINDOW", True);
        XSetWMProtocols(GLWin.dpy, GLWin.win, &wmDelete, 1);
        XSetStandardProperties(GLWin.dpy, GLWin.win, title,
            title, None, NULL, 0, NULL);
        XMapRaised(GLWin.dpy, GLWin.win);
    }       
    /* connect the glx-context to the window */
    glXMakeCurrent(GLWin.dpy, GLWin.win, GLWin.ctx);
    XGetGeometry(GLWin.dpy, GLWin.win, &winDummy, &GLWin.x, &GLWin.y,
        &GLWin.width, &GLWin.height, &borderDummy, &GLWin.depth);
    printf("Depth %d\n", GLWin.depth);
    printf("%dx%d\n", GLWin.width, GLWin.height);
    if (glXIsDirect(GLWin.dpy, GLWin.ctx)) 
        printf("Congrats, you have Direct Rendering!\n");
    else
        printf("Sorry, no Direct Rendering possible!\n");
    initGL();
    return True;    
}

GLvoid glPrintf(int font, const char * fmt, ...)
{
    va_list ap;
    char text[256];
    if (fmt == NULL)
        return;

    va_start(ap, fmt);
    vsprintf(text, fmt, ap);
    va_end(ap);
    glPushAttrib(GL_LIST_BIT);
    glListBase(base + font*96 - 32);
    glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
    glPopAttrib();
}

int serial_open(char * dev)
{
    struct termios tio;
    int fd;
    char * dbg_on = "DEBUG ON\n";
    char * ln_on = "LOCALIZATION ON\n";
    char * cn_on = "CONES ON\n";

    fprintf(stderr, "Using %s\n", dev);
                                                                                
    fd = open(dev, O_RDWR);
    if (fd == -1) {
        perror("open");
        return -1;
    }
                                                                                
    /* Set the serial port attributes */
    tcgetattr(fd, &tio);
                                                                                
    tio.c_cflag = CS8 | CLOCAL | CREAD;
    tio.c_iflag = IGNPAR | IGNCR;
    cfmakeraw(&tio);
                                                                                
    cfsetospeed(&tio, B115200);
    cfsetispeed(&tio, B115200);
                                                                                
    tcflush(fd, TCIFLUSH);
    if (tcsetattr(fd, TCSANOW, &tio) == -1) {
        perror("tcsetattr");
        return -1;
    }

    write(fd, dbg_on, strlen(dbg_on));
    usleep(300000);
    write(fd, ln_on, strlen(ln_on));
    if (do_cones) {
        usleep(300000);
        write(fd, cn_on, strlen(cn_on));
    }

    return fd;
}

int node_find_v1(int id)
{
    int i;
    for (i=0; i<=max_nodes; i++) {
        if ((pos[i].stat & STAT_VALID) && pos[i].id[3] == id)
            return i;
    }
    return -1;
}

int id_match(int * id1, int * id2)
{
    int i;
    for (i=0; i<4; i++) {
        if (id1[i] != id2[i])
            return 0;
    }
    return 1;
}

void node_new(int i, int * id)
{
    int j;
    char buf[8];
    pos[i].stat = STAT_VALID;
    pos[i].newstat = pos[i].stat;
    pos[i].id[0] = id[0];
    pos[i].id[1] = id[1];
    pos[i].id[2] = id[2];
    pos[i].id[3] = id[3];
    pos[i].x = 0;
    pos[i].y = 0;
    pos[i].last_x = 0;
    pos[i].last_y = 0;
    pos[i].hist = NULL;
    pos[i].hist_size = 0;
    pos[i].hist_num = 0;
    pos[i].vel_set = 0;
    for (j=0; j<=max_nodes; j++) {
        pings[i][j].tv_sec = 0;
        pings[j][i].tv_sec = 0;
    }
    for (j=0; j<NUM_NAMES; j++) {
        if (id_match(names[j].id, id)) {
            pos[i].name = names[j].name;
            printf("# node %d is called \"%s\"\n", i, names[j].name);
            break;
        }
    }
    if (j == NUM_NAMES) {
        sprintf(buf, "0x%02x", id[3]);
        pos[i].name = strdup(buf);
    }

    if (do_flip && !strcmp(flipname, pos[i].name)) {
        do_flip = 2;
        flip_node = i;
        printf("Flip node is %d\n", i);
    }
    if (do_rot && !strcmp(rotname, pos[i].name)) {
        do_rot = 2;
        rot_node = i;
        printf("Rot node is %d\n", i);
    }
}

void print_matrix(float * m, int r, int c)
{
    int i, j;
    printf("[ ");
    for (i=0; i<r; i++) {
        for (j=0; j<c; j++) {
            printf("%f ", m[j*r+i]);
        }
        printf("; ");
    }
    printf("]\n");
}


void matrix_mult(float * dst, float * m1, float * m2, int r1, int c1, int c2)
{
    int i, j, k;
    //printf("srcs mult: ");
    //print_matrix(m1, r1, c1);
    //print_matrix(m2, c1, c2);
    for (i=0; i<c2; i++) {
        for (j=0; j<r1; j++) {
            float * m = m1+j;
            *dst = 0;
            for (k=0; k<c1; k++) {
                *dst += m2[k] * m[k*r1];
            }
            dst++;
        }
        m2 += c1;
    }
    //printf("dst ");
    //print_matrix(dst, r1, c2);
}

void matrix_mult_trans_2(float * dst, float * m1, float * m2, int r1, int c1, int r2)
{
    int i, j, k;
    //printf("srcs mult_trans: ");
    //print_matrix(m1, r1, c1);
    //print_matrix(m2, r2, c1);
    for (i=0; i<r2; i++) {
        for (j=0; j<r1; j++) {
            float * a = m1+j;
            float * b = m2+i;
            *dst = 0;
            for (k=0; k<c1; k++) {
                *dst += b[k*r2] * a[k*r1];
            }
            dst++;
        }
    }
    //printf("dst ");
    //print_matrix(dst, r1, r2);
}

void matrix_inv_2x2(float * dst, float * m)
{
    float div = 1.0 / (m[0]*m[3] - m[2]*m[1]);
    dst[0] = m[3] * div;
    dst[1] = -m[1] * div;
    dst[2] = -m[2] * div;
    dst[3] = m[0] * div;
}

void update_vels(int i, char * str) {
    struct node_pos * p = pos+i;
    int j, n;
    float vs[max_nodes];
    float xy[max_nodes*2];
    float B[4];
    float C[4];
    float v[2];

    if (p->vel_set && do_smoothing) {
        float t;
        struct timeval tv;
        getcrickettime(&tv);
        t = sec_diff(tv, p->vel_time);
        p->last_x += p->vel_x * t;
        p->last_y += p->vel_y * t;
        p->x = p->last_x;
        p->y = p->last_y;
        //printf("new xy %f %f t %f\n", p->x, p->y, t);
    }
    n = 0;
    while (1) {
        float xdif, ydif, dist;
        int ret, vel, c;
        ret = sscanf(str, "%d v=%d;%n", &j, &vel, &c);
        if (ret != 2)
            break;
        str += c;
        if (!(pos[j].stat & STAT_VISIBLE))
            continue;
        if (vel > 1000 || vel < -1000)
            continue;
        vs[n] = vel / 30.0;
        xdif = pos[i].x - pos[j].x;
        ydif = pos[i].y - pos[j].y;
        //printf("xy %f %f xy %f %f\n", pos[i].x, pos[i].y, pos[j].x, pos[j].y);
        dist = sqrt(xdif*xdif + ydif*ydif);
        if (dist == 0.0)
            continue;
        //printf("xdif %f ydif %f dist %f\n", xdif, ydif, dist);
        xy[n<<1]     = xdif / dist;
        xy[(n<<1)+1] = ydif / dist;
        n++;
    }
    if (n < 2) {
        p->vel_set = 0;
        return;
    }
    matrix_mult_trans_2(B, xy, xy, 2, n, 2);
    matrix_inv_2x2(C, B);
    matrix_mult(B, xy, vs, 2, n, 1);
    matrix_mult(v, C, B, 2, 2, 1);
    p->vel_set = 1;
    p->vel_x = v[0];
    p->vel_y = v[1];
    //printf("vel is %f %f\n", v[0], v[1]);
    getcrickettime(&p->vel_time);
}

void update_history(int i)
{
    struct node_pos * p = pos+i;

    if (!(p->stat & STAT_MOVING)) {
        if (p->hist) {
            free(p->hist);
            p->hist_num = 0;
            p->hist_size = 0;
        }
        return;
    }

    if (p->hist_num == p->hist_size) {
        p->hist_size += 20;
        p->hist = realloc(p->hist, p->hist_size * sizeof(struct hist_pos));
    }

    p->hist[p->hist_num].x = p->x;
    p->hist[p->hist_num].y = p->y;
    getcrickettime(&p->hist[p->hist_num].time);
    p->hist_num++;
}

int node_new_v1(int id)
{
    int i, j;
    for (i=0; i<max_nodes; i++) {
        if (pos[i].stat == STAT_EMPTY) {
            pos[i].stat = STAT_VALID;
            pos[i].id[3] = id;
            for (j=0; j<=max_nodes; j++) {
                pings[i][j].tv_sec = 0;
                pings[j][i].tv_sec = 0;
            }
            pos[i].name = "Unk";
            for (j=0; j<NUM_NAMES; j++) {
                if (names[j].crc == id)
                    pos[i].name = names[j].name;
            }
            return i;
        }
    }
    return -1;
}

void update_mob_dist(char * str)
{
    int j, c;
    int ret, dist;

    while (1) {
        ret = sscanf(str, " %d d=%d ;%n", &j, &dist, &c);
        str += c;
        if (ret != 2)
            break;
        if (do_cones) {
            dist -= 180;
        }
        mob_dist_new[j] = dist;
    }
}

void process_dists_v2(int node, char * str)
{
    int j, c;
    int ret, dist;

    while (1) {
        ret = sscanf(str, " %d d=%d ;%n", &j, &dist, &c);
        str += c;
        if (ret != 2)
            break;
//        getcrickettime(&pings[node][j]);
        if (do_cones) {
            dist -= 180;
        }
        pings_summary_new[node][j] = dist;
        if (input_version == 3)
            pings_summary_new[j][node] = dist;
        if (edge_picked && node != max_nodes && node == edge_v[0] &&
                j == edge_v[1] && dist != pings_summary[node][j]) {
            new_edge = 1;
            new_edge_len = dist/(float)30;
        }
//        printf("%02d %02x %05d\n", node, id, dist);
    }
}

void process_dists_v1(int node, char * str)
{
    int j, c;
    int id, dist;

    while (1) {
        j = sscanf(str, " 0x%x d=%d ;%n", &id, &dist, &c);
        str += c;
        if (j != 2)
            break;
        j = node_find_v1(id);
        if (j == -1)
            continue;
//        getcrickettime(&pings[node][j]);
        pings_summary_new[node][j] = dist;
        if (edge_picked && node != max_nodes && node == edge_v[0] &&
                j == edge_v[1] && dist != pings_summary[node][j]) {
            new_edge = 1;
            new_edge_len = dist/(float)30;
        }
//        printf("%02d %02x %05d\n", node, id, dist);
    }
}

int solve_read = 0;
int opt_read = 0;
int last_node;

void process_line_v2(char * str)
{
    int i;
    int ret;
    float x, y;
    struct timeval tv;
    
    getcrickettime(&tv);
    printf("%d.%03d: %s\n", tv.tv_sec, tv.tv_usec/1000, str);

    if (!strncmp(str, "Localized ", 10)) {
        solve_read = 1;
        for (i=0; i<max_nodes; i++) {
            if (pos[i].stat & STAT_VALID) {
#if 0
                if (pos[i].stat & STAT_MARKED) {
                    pos[i].stat &= ~STAT_VISIBLE;
                    pos[i].stat &= ~STAT_OPTIMIZED;
                }
                pos[i].stat |= STAT_MARKED;
#endif
                pos[i].newstat = pos[i].stat;
                if (!(pos[i].stat & STAT_MOVING))
                    pos[i].newstat &= ~STAT_VISIBLE;
                pos[i].newstat &= ~STAT_OPTIMIZED;
            }
        }
    }
    else if (!strncmp(str, "Optimized,", 10)) {
        float err;
        ret = sscanf(str, "Optimized, sumsq=%f", &err);
        if (ret == 1) {
            opt_err = err;
        }
        opt_read = 1;
        solve_read = 0;
    }
    else if (!strcmp(str, "Localizing...")) {
        num_quads_new = 0;
    }
    else if (!strncmp(str, "Rigid Quad:", 11)) {
        int ids[3];
        sscanf(str, "Rigid Quad: %d %d %d", ids, ids+1, ids+2);
        for (i=0; i<3; i++)
            quads_new[num_quads_new].v[i] = ids[i];
        num_quads_new++;
    }
    else if (!strncmp(str, "RQ:", 3)) {
        int ids[3];
        sscanf(str, "RQ: %d %d %d", ids, ids+1, ids+2);
        for (i=0; i<3; i++)
            quads_new[num_quads_new].v[i] = ids[i];
        num_quads_new++;
    }
    else if (sscanf(str, "  %d: x=%f y=%f", &i, &x, &y) == 3) {
        if (solve_read) {
            pos[i].newstat |= STAT_VISIBLE;
            //pos[i].stat &= ~STAT_MARKED;
            if (strchr(str, '*'))
                pos[i].newstat |= STAT_WEAK;
            else
                pos[i].newstat &= ~STAT_WEAK;
            if (strchr(str, 'm')) {
                pos[i].newstat |= STAT_MOVING;
                pos[i].ox = x;
                pos[i].oy = y;
            }
            else {
                pos[i].newstat &= ~STAT_MOVING;
                pos[i].x = x;
                pos[i].y = y;
                pos[i].last_x = x;
                pos[i].last_y = y;
                getcrickettime(&pos[i].vel_time);
            }
        }
        else if (opt_read) {
           /* if (opt_err > 1000000) {
                x = pos[i].ox;
                y = pos[i].oy;
            }
            else {
                pos[i].newstat |= STAT_OPTIMIZED;
            }
            */
            if (opt_err < 1000000) {
                pos[i].x = x;
                pos[i].y = y;
                pos[i].last_x = x;
                pos[i].last_y = y;
                getcrickettime(&pos[i].vel_time);
                update_history(i);
                memcpy(mob_dist[i], mob_dist_new, sizeof(mob_dist_new));
            }
        }
    }
    else if (!strcmp(str, "done")) {
        opt_read = solve_read = 0;
        memcpy(quads, quads_new, sizeof(quads));
        num_quads = num_quads_new;
        memcpy(pings_summary, pings_summary_new, sizeof(pings_summary));
        for (i=0; i<max_nodes; i++) {
            if (pos[i].stat & STAT_VALID) {
                pos[i].stat = pos[i].newstat;
            }
        }
        //drawGLScene();
    }
    else if (!strcmp(str, "opt done")) {
        opt_read = 0;
        if (pause_on_loc)
            paused = 1;
    }
    else if (!strncmp(str, "beacon:", 7)) {
        int dist, time;
        ret = sscanf(str, "beacon:%d,dist=%d,time=%d", &i, &dist, &time);
        if (ret != 3)
            return;
        if (i == 0xff)
            return;
        getcrickettime(&pings[max_nodes][i]);
        if (time/dist == 28)
            pings_d[max_nodes][i] = time * 1.08;
        else
            pings_d[max_nodes][i] = time;
        if (do_cones)
            pings_d[max_nodes][i] -= 180;
        if (edge_picked && edge_v[0] == max_nodes && edge_v[1] == i) {
            new_edge = 1;
            new_edge_len = pings_d[max_nodes][i]/(float)30;
        }
        //drawGLScene();
    }
    else if (!strncmp(str, "node ", 5)) {
        int id[4];
        char info[64];
        info[0] = '\0';
        if (sscanf(str, "node %d (self) ID %x:%x:%x:%x %64s",
                    &i, id, id+1, id+2, id+3, info) >= 5) {
            max_nodes = i;
            node_new(max_nodes, id);
            pos[max_nodes].stat |= STAT_VISIBLE;
        }
        else if (sscanf(str, "node %d ID %x:%x:%x:%x %64s",
                    &i, id, id+1, id+2, id+3, info) >= 5) {
            node_new(i, id);
        }
    }
    else if (!strncmp(str, "  from ", 7)) {
        char buf[16];
        ret = sscanf(str, " from %d%7c:", &i, buf);
        buf[7] = '\0';
        if (ret == 2) {
            if (summary_complete) {
                bzero(pings_summary_new, sizeof(pings_summary_new));
                summary_complete = 0;
            }
            if (!strcmp(buf, " (self)")) {
                summary_complete = 1;
            }
            last_node = i;
            process_dists_v2(i, str+17);
        }
    }
    else if (!strncmp(str, "          ", 10)) {
        process_dists_v2(last_node, str);
    }
    else if (!strncmp(str, "  vels to", 9)) {
        int c;
        ret = sscanf(str, " vels to %d:%n", &i, &c);
        if (ret != 1)
            return;
        str += c;
        update_vels(i, str);
    }
    else if (!strncmp(str, "  dist", 6)) {
        bzero(mob_dist_new, sizeof(mob_dist_new));
        update_mob_dist(str+6);
    }
    else if (!strncmp(str, "Pred:", 5)) {
        int dist;
        ret = sscanf(str, "Pred:%d,dist=%d", &i, &dist);
        if (ret != 2)
            return;
        pings_summary_new[max_nodes][i] = dist;
    }
}

void process_line_v1(char * str)
{
    int i;
    int id;
    float x, y;
    struct timeval tv;
    
    getcrickettime(&tv);
    printf("%d.%03d: %s\n", tv.tv_sec, tv.tv_usec/1000, str);

    if (!strncmp(str, "Localized ", 10)) {
        solve_read = 1;
        for (i=0; i<max_nodes; i++) {
            if (pos[i].stat & STAT_VALID) {
                if (pos[i].stat & STAT_MARKED) {
                    pos[i].stat = STAT_EMPTY;
                }
                else {
                    pos[i].stat |= STAT_MARKED;
                }
            }
        }
    }
    else if (!strncmp(str, "Optimized,", 10)) {
        opt_read = 1;
    }
    else if (!strcmp(str, "Localizing...")) {
        num_quads_new = 0;
    }
    else if (!strncmp(str, "Rigid Quad:", 11)) {
        int ids[3];
        sscanf(str, "Rigid Quad: %x %x %x", ids, ids+1, ids+2);
        for (i=0; i<3; i++)
            quads_new[num_quads_new].v[i] = node_find_v1(ids[i]);
        num_quads_new++;
    }
    else if (!strncmp(str, "  0x", 4)) {
        sscanf(str, "  0x%x: x=%f y=%f", &id, &x, &y);
        i = node_find_v1(id);
        if (i == -1)
            i = node_new_v1(id);

        if (i != -1) {
            if (solve_read) {
                pos[i].x = x;
                pos[i].y = y;
                pos[i].last_x = x;
                pos[i].last_y = y;
                pos[i].stat |= STAT_VISIBLE;
                pos[i].stat &= ~STAT_MARKED;
            }
            else if (opt_read) {
                pos[i].ox = x;
                pos[i].oy = y;
                pos[i].stat |= STAT_OPTIMIZED;
            }
        }
    }
    else if (!strcmp(str, "done")) {
        solve_read = 0;
        memcpy(quads, quads_new, sizeof(quads));
        num_quads = num_quads_new;
        memcpy(pings_summary, pings_summary_new, sizeof(pings_summary));
        //drawGLScene();
    }
    else if (!strcmp(str, "opt done")) {
        opt_read = 0;
    }
    else if (!strncmp(str, "id=", 3)) {
        int dist, time;
        int myid, j;
        i = sscanf(str, "id=%x,beacon=%x,dist=%d,time=%d", &myid, &id, &dist, &time);
        if (i != 4)
            return;
        for (j=0; j<NUM_NAMES; j++) {
            if (names[j].crc == myid) {
                pos[max_nodes].name = names[j].name;
                pos[max_nodes].id[3] = myid;
            }
        }
        i = node_find_v1(id);
        if (i == -1)
            return;
        getcrickettime(&pings[max_nodes][i]);
        pings_d[max_nodes][i] = time;
        if (edge_picked && edge_v[0] == max_nodes && edge_v[1] == i) {
            new_edge = 1;
            new_edge_len = time/(float)30;
        }
        //drawGLScene();
    }
    else if (!strncmp(str, "  from 0x", 9)) {
        char buf[16];
        i = sscanf(str, " from 0x%x%7c:", &id, buf);
        buf[7] = '\0';
        if (i == 2) {
            if (summary_complete) {
                bzero(pings_summary_new, sizeof(pings_summary_new));
                summary_complete = 0;
            }
            if (!strcmp(buf, " (self)")) {
                summary_complete = 1;
            }
            i = node_find_v1(id);
            if (i == -1)
                return;
            last_node = i;
            process_dists_v1(i, str+19);
        }
    }
    else if (!strncmp(str, "          0x", 12)) {
        process_dists_v1(last_node, str);
    }
}

char linebuf[256];

void get_serial_input(int serfd)
{
    struct timeval tv;
    fd_set rfds;
    char buf[256];
    int len;
    char * a, * s;

    tv.tv_sec = 0;
    tv.tv_usec = 50000;
    FD_ZERO(&rfds);
    FD_SET(serfd, &rfds);

    if (select(serfd+1, &rfds, NULL, NULL, &tv) == 1) {
        len = read(serfd, buf, sizeof(buf)-1);
        buf[len] = '\0';
        s = buf;
        while ((a = strchr(s, '\n'))) {
            *a = '\0';
            strcat(linebuf, s);
            process_line(linebuf);
            strcpy(linebuf, "");
            s = a+1;
        }
        strcat(linebuf, s);
    }
}

struct timeval tdiff = {0, 0};

void settimediff(int skipusec)
{
    struct timeval tv;
    int lower;
    gettimeofday(&tv, NULL);
    /*
    if (tv.tv_usec < crickettime.tv_usec) {
        tdiff.tv_sec = tv.tv_sec - crickettime.tv_sec - 1;
        tdiff.tv_usec = 1000000 + tv.tv_usec - crickettime.tv_usec;
    }
    else {
        tdiff.tv_sec = tv.tv_sec - crickettime.tv_sec;
        tdiff.tv_usec = tv.tv_usec - crickettime.tv_usec;
    }
    */

    lower = tv.tv_usec-crickettime.tv_usec-skipusec;
    tdiff.tv_sec = tv.tv_sec - crickettime.tv_sec + (lower-1000000)/1000000;
    lower %= 1000000;
    tdiff.tv_usec = (lower < 0)?(1000000+lower):lower;
}


int get_file_input(FILE * in, int wait)
{
    struct timeval tv;
    int sleeptime = 50000;
    int c, sec, usec;

    gettimeofday(&tv, NULL);
    do {
        c = sscanf(linebuf, "%d.%d:", &sec, &usec);
        if (c != 2)
            continue;
        usec *= 1000;
        if (tdiff.tv_sec == 0 && tdiff.tv_usec == 0) {
            crickettime.tv_sec = sec;
            crickettime.tv_usec = usec;
            settimediff(0);
        }
        sleeptime = (sec + tdiff.tv_sec - tv.tv_sec) * 1000000 +
            (usec + tdiff.tv_usec - tv.tv_usec);
        if (sleeptime < 0)
            sleeptime = 0;

        crickettime.tv_sec = sec + (usec-sleeptime-1000000)/1000000;
        crickettime.tv_usec = (usec+1000000-sleeptime)%1000000;
        
        if (sleeptime > 50000)
            sleeptime = 50000;

        if (sleeptime > 0) {
            if (wait)
                usleep(sleeptime);
     //                   getcrickettime(&tv);
     //                   printf("s = %d, us = %d\n", tv.tv_sec, tv.tv_usec);
            return 0;
        }
        linebuf[strlen(linebuf)-1] = '\0';
        process_line(strchr(linebuf,' ')+1);
    } while (fgets(linebuf, sizeof(linebuf), in));

    strcpy(linebuf, "");
    if (wait)
        usleep(50000);
    return -1;
}

void read_truth(char * filename)
{
    FILE * in;
    char buf[256];

    in = fopen(filename, "r");
    if (!in)
        return;

    free(truth);
    num_truth = 0;

    while (fgets(buf, sizeof(buf), in)) {
        char name[8];
        float x,y;
        if (sscanf(buf, "%8s %f %f\n", name, &x, &y) != 3)
            continue;

        truth = realloc(truth, (num_truth+1)*sizeof(struct truth_pos));
        truth[num_truth].x = x;
        truth[num_truth].y = y;
        truth[num_truth].name = strdup(name);
        num_truth++;
    }

    fclose(in);
}

int main(int argc, char **argv)
{
    int fd;
    XEvent event;
    Bool done;
    char c;
    FILE * input = NULL;
    char outstr[256];
    int do_output = 0;
    int frame = 0;

    while ((c = getopt(argc, argv, "i:v:t:o:cr:f:s")) >= 0) {
        switch (c) {
            case 'i':
                input = fopen(optarg, "r");
                isrecorded = 1;
                if (!input) {
                    perror("open");
                    exit(1);
                }
                break;
            case 'v':
                input_version = atoi(optarg);
                if (input_version == 1) {
                    process_line = process_line_v1;
                    max_nodes = 20;
                }
                break;
            case 't':
                read_truth(optarg);
                break;
            case 'o':
                do_output = 1;
                paused = 1;
                strcpy(outstr, optarg);
                break;
            case 'c':
                do_cones = 1;
                break;
            case 'r':
                do_rot = 1;
                strcpy(rotname, optarg);
                break;
            case 'f':
                do_flip = 1;
                strcpy(flipname, optarg);
                break;
            case 's':
                do_smoothing = 1;
                break;
            default:
                exit(1);
        }
    }

    if (do_rot != do_flip) {
        fprintf(stderr, "Must specify -r and -f together\n");
        exit(1);
    }
    
    strcpy(linebuf, "");
    init_pos();

    if (!input) {
        fd = serial_open("/dev/ttyS0");
        if (fd < 0) {
            return 1;
        }
    }

    done = False;
    /* default to fullscreen */
    GLWin.fs = False;
    createGLWindow("Cricket Viewer", WIDTH, HEIGHT, 24, GLWin.fs);

    if (input)
        get_file_input(input, 0);

    getcrickettime(&starttime);

    /* wait for events*/ 
    while (!done)
    {
        if (input) {
            if (!paused)
                get_file_input(input, 1);
            else if (do_output) {
                if (get_file_input(input, 0) == -1)
                    break;
            }
            else
                usleep(50000);
        }
        else
            get_serial_input(fd);

        if (auto_zoom) {
            recompute_bounding(&target_ox, &target_oy, &target_fov, origin_x,
                    origin_y, fov);

            if (do_rot == 2 && (pos[rot_node].stat & STAT_VISIBLE)) {
                float newrot;
                if (do_flip == 2 && pos[flip_node].stat & STAT_VISIBLE) {
                    float cross = pos[flip_node].x * pos[rot_node].y - pos[flip_node].y * pos[rot_node].x;
                    /* last_cross will be negative if cross and last_cross
                     * differ in sign */
                    last_cross *= cross;
                    if (last_cross < 0) {
                        comp_flip = -comp_flip;
                        printf("adjust: FLIPPED\n");
                    }
                    last_cross = cross;
                }
                newrot = -atan2(pos[rot_node].y, comp_flip*pos[rot_node].x);
                newrot = 180 * newrot / M_PI;
                comp_rot += (newrot - last_rot);
                printf("adjust: %f\n", comp_rot);
                last_rot = newrot;
            }

            origin_x += (target_ox - origin_x) * 0.1;
            origin_y += (target_oy - origin_y) * 0.1;
            fov += (target_fov - fov) * 0.1;
        }
        /* handle the events in the queue */
        while (XPending(GLWin.dpy) > 0)
        {
            KeySym k;
            XNextEvent(GLWin.dpy, &event);
            switch (event.type)
            {
                case Expose:
	                if (event.xexpose.count != 0)
	                    break;
                    drawGLScene();
         	        break;
            case ConfigureNotify:
            /* call resizeGLScene only if our window-size changed */
                if ((event.xconfigure.width != GLWin.width) || 
                    (event.xconfigure.height != GLWin.height))
                {
                    GLWin.width = event.xconfigure.width;
                    GLWin.height = event.xconfigure.height;
                    //    printf("Resize event\n");
                    resizeGLScene(event.xconfigure.width,
                        event.xconfigure.height);
                }
                break;
            case ButtonPress:
                if (picking_edge == 1) {
                    edge_v[0] = find_node_by_xy(event.xbutton.x, event.xbutton.y);
                    picking_edge = 2;
                }
                else if (picking_edge == 2) {
                    struct timeval tv;
                    edge_v[1] = find_node_by_xy(event.xbutton.x, event.xbutton.y);
                    picking_edge = 0;
                    edge_picked = 1;
                    edge_head = 0;
                    edge_tail = 0;
                    getcrickettime(&tv);
                    edgeupdate.tv_sec = tv.tv_sec;
                    edgeupdate.tv_usec = tv.tv_usec;
                }
                else {
                    show_vertex = find_node_by_xy(event.xbutton.x, event.xbutton.y);
                    if (show_vertex != -1)
                        print_rigidity(show_vertex);
                }
                break;
            case KeyPress:
                k = XLookupKeysym(&event.xkey, 0);
                if (k == XK_Escape)
                {
                    printf("quitting...\n");
                    done = True;
                }
                if (k == XK_F1)
                {
                    killGLWindow();
                    GLWin.fs = !GLWin.fs;
                    createGLWindow("Cricket Viewer", WIDTH, HEIGHT, 24, GLWin.fs);
                }
                if (k == XK_Left) {
                    origin_x -= 2;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_Right) {
                    origin_x += 2;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_Down) {
                    origin_y -= 2;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_Up) {
                    origin_y += 2;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_Page_Up) {
                    fov += 5;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_Page_Down) {
                    fov -= 5;
                    auto_zoom = 0;
                    drawGLScene();
                }
                if (k == XK_i) {
                    trot += 0.5;
                    drawGLScene();
                }
                if (k == XK_o) {
                    trot -= 0.5;
                    drawGLScene();
                }
                if (k == XK_k) {
                    trot += 5;
                    drawGLScene();
                }
                if (k == XK_l) {
                    trot -= 5;
                    drawGLScene();
                }
                if (k == XK_comma) {
                    ui_rot += 5;
                    drawGLScene();
                }
                if (k == XK_period) {
                    ui_rot -= 5;
                    drawGLScene();
                }
                if (k == XK_f) {
                    ui_flip = -ui_flip;
                    drawGLScene();
                }
                if (k == XK_b) {
                //    target_fov = 1;
                    auto_zoom = 1;
                //    recompute_bounding(&target_ox, &target_oy, &target_fov);
                }
                if (k == XK_p) {
                    show_pings = (show_pings+1)%3;
                }
                if (k == XK_q) {
                    show_quads = !show_quads;
                }
                if (k == XK_h) {
                    show_all_history = !show_all_history;
                }
                if (k == XK_w) {
                    if (picking_edge)
                        picking_edge = 0;
                    else {
                        picking_edge = 1;
                        edge_picked = 0;
                    }
                }
                if (k == XK_c) {
                    write_frame("outframe.png");
                }
                if (k == XK_m) {
                    struct timeval tv;
                    getcrickettime(&tv);
                    printf("%d.%03d: --MARK--\n", tv.tv_sec, tv.tv_usec/1000);
                }
                if (input) {
                    if (k == XK_space) {
                        if (!paused) {
                            paused = 1;
                        }
                        else {
                            paused = 0;
                            settimediff(0);
                            pause_on_loc = 0;
                        }
                    }
                    if (k == XK_n) {
                        pause_on_loc = 1;
                        if (paused) {
                            paused = 0;
                            settimediff(0);
                        }
                    }
                    if (k == XK_s) {
                        settimediff(100000);
                        if (paused)
                            get_file_input(input, 1);
                    }
                    if (k == XK_a) {
                        long pos;
                        settimediff(-100000);
                        strcpy(linebuf, "");
                        init_pos();
                        pos = ftell(input);
                        pos -= 5000;
                        if (pos < 0)
                            pos = 0;
                        fseek(input, pos, SEEK_SET);

                        if (paused)
                            get_file_input(input, 1);
                    }
                }
                    
                break;
            case ClientMessage:    
                if (*XGetAtomName(GLWin.dpy, event.xclient.message_type) == 
                    *"WM_PROTOCOLS")
                {
                    printf("Exiting sanely...\n");
                    done = True;
                }
                break;
            default:
                break;
            }
        }
        drawGLScene();
        if (do_output) {
            char filename[256];
            snprintf(filename, sizeof(filename), outstr, frame);
            write_frame(filename);
            settimediff(33367);
            frame++;
        }
    }
    killGLWindow();

    if (input)
        fclose(input);
    else {
        char * dbg_off = "DEBUG OFF\n";
        char * ln_off = "LOCALIZATION OFF\n";
        char * cn_off = "CONES OFF\n";
        write(fd, ln_off, strlen(ln_off));
        usleep(300000);
        write(fd, dbg_off, strlen(dbg_off));
        if (do_cones) {
            usleep(300000);
            write(fd, cn_off, strlen(cn_off));
        }

        tcflush(fd, TCIOFLUSH);
        close(fd);
    }

    return 0;
}
